#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for H/XR/XV Migration."""


from __future__ import with_statement

import os
import urllib
import urllib2
import cookielib
import re

from HTMLParser import (HTMLParser,)

from iobase.exception import (InvalidArgError, ExecError, IoBaseException)
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.xmldata import XmlConfNode
from iobase.fileut import (FLock,)
from iobase import netut
from product_conf import (FIND_LANDISK_CMD, PING_CMD, PING6_CMD, HOSTNAME_CMD)
from pal_info import (ReqTypes, Results)
from albg_info import (ALBG_MIGRATION_PATH,)

import pal_config as config
import libpal.palcmd as palcmd
import libpal.crond.ctrl as crond_ctrl

from pkgaddon.pal_hmigration_config import (GET_REMOTE_LANDISK_INFO_CMD,
                                            START_FILE_PATH, STATUS_FILE_PATH,
                                            PROGRESS_FILE_PATH,
                                            STATUS_NO_EXEC, STATUS_PROCESSING,
                                            STATUS_ABORT, EXEC_LOCK_NAME,
                                            EXEC_MIGRATION_CMD,)


MIGRATION_NAME = "migration"
SOURCELIST_NAME = "sourcelist"
SOURCE_NAME = "source"
NAME_NAME = "name"
NAME_PATH = "source/name"
PASSWD_PATH = "source/passwd"
LANDISKINFO_NAME = "landiskinfo"
JOB_NAME = "job"
HOSTNAMELIST_NAME = "hostnamelist"
HOSTNAME_PATH = "hostnamelist/hostname"
NAME_IN_JOB_PATH = "job/sourcelist/source/name"
PASSWD_IN_JOB_PATH = "job/sourcelist/source/passwd"
SCHEDULE_PATH = "job/schedule"
YEAR_PATH = "date/year"
MONTH_PATH = "date/month"
DAY_PATH = "date/day"
HOUR_PATH = "date/hour"
MINUTE_PATH = "date/minute"
PRODUCT_NAME_PATH = "product/name"

STARTTIME_NAME = "starttime"
ENDTIME_NAME = "endtime"
TIMESTAMP_NAME = "timestamp"
ELAPSEDTIME_NAME = "elapsedtime"
PROGRESS_NAME = "progress"

STATUSINFO_NO_EXEC_XML = "<statusinfo><status>NO_EXEC</status></statusinfo>"
STATUSINFO_XML_TMPL = """<statusinfo>
    <status>%s</status>
    <starttime/>
    <endtime/>
    <elapsedtime/>
    <progress/>
</statusinfo>
"""

XR_XV_PRODUCT_PREFIX_LIST = ("HDL-XR", "HDL-XV")


PING_ARG_BASE = [PING_CMD, "-c", "1", "-W", "1"]
PING6_ARG_BASE = [PING6_CMD, "-c", "1", "-W", "1"]

CRON_ENTRY_NAME = "migration"
CRON_USER = "root"
CRON_DEFAULT = "*"

EXIT_CODE_AUTH_ERROR = 6


LOGON_CGI_URL_TMPL = "http://%s/logon.cgi"
LOGOFF_CGI_URL_TMPL = "http://%s/logoff.cgi"
DATETIME_CGI_URL_TMPL = "http://%s/detail/basic/datetime/datetime.cgi"
PROXY_CGI_URL_TMPL = "http://%s/detail/basic/proxy/proxy.cgi"
SNMP_CGI_URL_TMPL = "http://%s/detail/notice/snmp/snmp.cgi"

TRUE_TEXT = XmlConfNode.TRUE_TEXT
FALSE_TEXT = XmlConfNode.FALSE_TEXT

INPUT_TAG_NAME = "input"
SELECT_TAG_NAME = "select"
OPTION_TAG_NAME = "option"
FORM_TAG_NAME = "form"
ID_ATTR = "id"
NAME_ATTR = "name"
VALUE_ATTR = "value"
TYPE_ATTR = "type"
CHECKED_ATTR = "checked"
SELECTED_ATTR = "selected"
ACTION_ATTR = "action"

LOGON_PAGE_NAME = "logon.cgi"
PASSWORD_TYPE = "password"

BR_TAG_RE = "<\s*br\s*/\s*>"

DATETIME_BASE_TMPL = """<datetime>
    <ntp enable="%(ntp_enable)s">
        <serveraddress>%(serveraddress)s</serveraddress>
        <sync>
            <boot enable="%(boot_enable)s"/>
            <time enable="%(time_enable)s">
                <hour>%(hour)s</hour>
                <minute>%(minute)s</minute>
            </time>
        </sync>
    </ntp>
</datetime>"""

PROXY_BASE_TMPL = """<proxy enable="%(proxy_enable)s">
    <manual enable="%(manual_enable)s">
        <host>%(host)s</host>
        <port>%(port)s</port>
    </manual>
    <pacfile enable="%(pacfile_enable)s">
        <url>%(url)s</url>
    </pacfile>
</proxy>"""

SNMP_BASE_TMPL = """<snmp>
    <trap enable="%(trap_enable)s">
        <port>%(port)s</port>
        <host>%(host)s</host>
    </trap>
</snmp>"""


class LogonParser(HTMLParser):
    """Class for parse logon page."""

    def __init__(self):
        """Constructor."""

        HTMLParser.__init__(self)
        self._is_logon_page = False
        self.attrs_map = {}

    def handle_starttag(self, tag, attrs):
        """Handle start tag."""

        attrs_dict = dict(attrs)

        if (tag == FORM_TAG_NAME):
            if attrs_dict.get(ACTION_ATTR).endswith(LOGON_PAGE_NAME):
                self._is_logon_page = True
        elif tag == INPUT_TAG_NAME:
            if NAME_ATTR in attrs_dict:
                self.attrs_map[attrs_dict[NAME_ATTR]] = attrs_dict

    def is_logon_page(self):
        """Check if parsed HTML is of the logon page."""

        return self._is_logon_page

    def get_form_data(self, passwd):
        """Return form data for logon."""

        form_data = {}
        for (name, attrs_dict) in self.attrs_map.items():
            if attrs_dict.get(TYPE_ATTR) == PASSWORD_TYPE:
                form_data[name] = passwd
            else:
                form_data[name] = attrs_dict.get(VALUE_ATTR)

        return form_data


class BaseHtmlParser(HTMLParser):
    """Base Class for parse HTML page."""

    def _to_attrs_dict(self, attrs):
        """Return a dictionary of attributes in "attrs"."""

        attrs_dict = {}
        for (key, value) in attrs:
            if (value != None and
                (value.lower() == TRUE_TEXT or value.lower() == FALSE_TEXT)):

                attrs_dict[key.lower()] = value.lower()
            else:
                attrs_dict[key.lower()] = value
        return attrs_dict

    def _enter_select(self, id_name):
        """Enter select tag whose id or name is "id_name"."""

        self.current_select_id_name = id_name

    def _exit_select(self):
        """Exit current select tag."""

        self.current_select_id_name = ""

    def _handle_input_tag(self, id_name, attrs_dict):
        """Handle input tag."""

        if id_name in self.target_id_name_list:
            self.attrs_map[id_name] = attrs_dict

    def _handle_option_tag(self, id_name, attrs_dict):
        """Handle option tag."""

        if id_name in self.target_id_name_list and SELECTED_ATTR in attrs_dict:
            self.attrs_map[id_name] = attrs_dict

    def __init__(self):
        """Constructor."""

        HTMLParser.__init__(self)
        self.attrs_map = {}
        self.current_select_id_name = ""
        self.target_id_name_list = []

    def handle_starttag(self, tag, attrs):
        """Handle start tag."""

        attrs_dict = self._to_attrs_dict(attrs)
        id_name = attrs_dict.get(ID_ATTR, "")
        if not id_name:
            id_name = attrs_dict.get(NAME_ATTR, "")

        if tag == INPUT_TAG_NAME:
            self._handle_input_tag(id_name, attrs_dict)
        elif tag == SELECT_TAG_NAME:
            self._enter_select(id_name)
        elif tag == OPTION_TAG_NAME:
            self._handle_option_tag(self.current_select_id_name, attrs_dict)

    def handle_endtag(self, tag):
        """Handle end tag."""

        if tag == SELECT_TAG_NAME:
            self._exit_select()

    def get_attrs(self, id_name):
        """Return a dictionary of attributes of tag whose id or name
           is "id_name".
        """

        return self.attrs_map.get(id_name, {})

    def get_attr(self, id_name, attr_name, default=""):
        """Return a value of "attr_name" attribute of tag whose id or name
           is "id".
        """

        return self.get_attrs(id_name).get(attr_name, default)

    def get_checked(self, id_name):
        """Check if there is "checked" attribute in attributes of tag
           whose id or name is "id_name" and return "true" or "false".
        """

        if CHECKED_ATTR in self.get_attrs(id_name):
            return TRUE_TEXT
        else:
            return FALSE_TEXT

    def get_xml_config(self):
        """Return XML configuration for target page."""

        raise NotImplemented


class DateTimeParser(BaseHtmlParser):
    """Class for parse datetime page."""

    NTP_ENABLE_ID = "sync_mode_1"
    NTP_SERVER_ID = "timeserver"
    BOOTSYNC_ID = "sync_timing__1"
    CRONSYNC_ID = "sync_timing__2"
    SYNC_HOUR_ID = "sync_every_hour"
    SYNC_MINUTE_ID = "sync_every_minute"

    def __init__(self):
        """Constructor."""

        BaseHtmlParser.__init__(self)

        self.target_id_name_list = [
                self.NTP_ENABLE_ID,
                self.NTP_SERVER_ID,
                self.BOOTSYNC_ID,
                self.CRONSYNC_ID,
                self.SYNC_HOUR_ID,
                self.SYNC_MINUTE_ID,
            ]

    def get_xml_config(self):
        """Return XML configuration for "datetime"."""

        if (not self.get_attr(self.SYNC_HOUR_ID, VALUE_ATTR) or
                not self.get_attr(self.SYNC_MINUTE_ID, VALUE_ATTR)):
            return None

        return XmlConfNode(DATETIME_BASE_TMPL %
                {
                    "ntp_enable": self.get_checked(self.NTP_ENABLE_ID),
                    "serveraddress":
                        self.get_attr(self.NTP_SERVER_ID, VALUE_ATTR),
                    "boot_enable": self.get_checked(self.BOOTSYNC_ID),
                    "time_enable": self.get_checked(self.CRONSYNC_ID),
                    "hour": self.get_attr(self.SYNC_HOUR_ID, VALUE_ATTR),
                    "minute": self.get_attr(self.SYNC_MINUTE_ID, VALUE_ATTR),
                })


class ProxyParser(BaseHtmlParser):
    """Class for parse proxy page."""

    PROXY_AUTO_ID = "proxy_mode_2"
    PROXY_MANUAL_ID = "proxy_mode_3"
    PROXY_SCRIPT_NAME = \
            "e887aae58b95e8a8ade5ae9ae382b9e382afe383aae38397e3838855524c"
    PROXY_HTTP_HOST_ID = "http_host"
    PROXY_HTTP_PORT_NAME = \
            "48545450e38397e383ade382ade382b75fe3839de383bce38388"

    def __init__(self):
        """Constructor."""

        BaseHtmlParser.__init__(self)

        self.target_id_name_list = [
                self.PROXY_AUTO_ID,
                self.PROXY_MANUAL_ID,
                self.PROXY_SCRIPT_NAME,
                self.PROXY_HTTP_HOST_ID,
                self.PROXY_HTTP_PORT_NAME,
            ]

    def get_xml_config(self):
        """Return XML configuration for "datetime"."""

        if not len(self.get_attrs(self.PROXY_AUTO_ID)):
            return None

        proxy_enable = FALSE_TEXT
        if (CHECKED_ATTR in self.get_attrs(self.PROXY_MANUAL_ID) or
            CHECKED_ATTR in self.get_attrs(self.PROXY_AUTO_ID)):

            proxy_enable = TRUE_TEXT

        return XmlConfNode(PROXY_BASE_TMPL %
                {
                    "proxy_enable": proxy_enable,
                    "manual_enable": self.get_checked(self.PROXY_MANUAL_ID),
                    "host": self.get_attr(self.PROXY_HTTP_HOST_ID, VALUE_ATTR),
                    "port":
                        self.get_attr(self.PROXY_HTTP_PORT_NAME, VALUE_ATTR),
                    "pacfile_enable": self.get_checked(self.PROXY_AUTO_ID),
                    "url": self.get_attr(self.PROXY_SCRIPT_NAME, VALUE_ATTR),
                })


class SnmpParser(BaseHtmlParser):
    """Class for parse snmp page."""

    TRAP_MODE_ID = "trap_mode_1"
    TRAP_HOST_NAME = \
            "e38388e383a9e38383e38397e98081e4bfa1e58588e3839be382b9e38388"
    TRAP_PORT_NAME = "e38388e383a9e38383e38397e98081e4bfa1e585" + \
                        "88e3839de383bce38388e795aae58fb7"

    def __init__(self):
        """Constructor."""

        BaseHtmlParser.__init__(self)

        self.target_id_name_list = [
                self.TRAP_MODE_ID,
                self.TRAP_HOST_NAME,
                self.TRAP_PORT_NAME,
            ]

    def get_xml_config(self):
        """Return XML configuration for "datetime"."""

        if not len(self.get_attrs(self.TRAP_MODE_ID)):
            return None

        return XmlConfNode(SNMP_BASE_TMPL %
                {
                    "trap_enable": self.get_checked(self.TRAP_MODE_ID),
                    "host": self.get_attr(self.TRAP_HOST_NAME, VALUE_ATTR),
                    "port": self.get_attr(self.TRAP_PORT_NAME, VALUE_ATTR),
                })


class RequestError(IoBaseException):
    """Request error class."""

    pass


class PalMigration(palcmd.PalCmd):
    """Pal Migration command implementation class."""

    #
    # Internal methods
    #

    def _check_body_node(self, node, type=None):
        """Check if entry's body node is valid."""

        if node.get_node_name() != MIGRATION_NAME:
            raise InvalidArgError(node.get_node_name())

        if type == ReqTypes.EXECUTE:
            try:
                node.get_node(LANDISKINFO_NAME)
            except:
                raise InvalidArgError("landiskinfo not found")

            try:
                node.get_node(JOB_NAME)
            except:
                raise InvalidArgError("job not found")

        if type == ReqTypes.CHECK:
            try:
                node.get_node(HOSTNAMELIST_NAME)
            except:
                raise InvalidArgError("landiskinfo not found")

        if type == ReqTypes.SWITCH:
            try:
                node.get_node(LANDISKINFO_NAME)
            except:
                raise InvalidArgError("landiskinfo not found")

    def _get_hostname(self):
        """Return current hostname."""

        proc = CmdExec(HOSTNAME_CMD, stdout=PIPE, stderr=NULL).execute()
        hostname = proc.stdout.read().strip()
        proc.wait()
        return hostname

    def _get_source_name_list(self):
        """Return source name list."""

        name_list = []
        proc = CmdExec(FIND_LANDISK_CMD, stdout=PIPE, stderr=NULL).execute()
        for line in proc.stdout.readlines():
            name_list.append(line.strip())
        proc.wait()

        return name_list

    def _is_hdl_xr_xv_series(self, landiskinfo):
        """Check if "landiskinfo" is of the HDL-XR/HDL-XV series."""

        product = landiskinfo.get_value(PRODUCT_NAME_PATH)
        for product_prefix in XR_XV_PRODUCT_PREFIX_LIST:
            if product.startswith(product_prefix):
                return True
        return False

    def _logon_and_get_html(self, opener, name, form_data):
        """Log on "name" host via HTTP and get HTML data in page specified
           in "form_data".
        """

        request = urllib2.Request(LOGON_CGI_URL_TMPL % (name,),
                                    urllib.urlencode(form_data))
        return opener.open(request).read()

    def _get_html(self, opener, url_tmpl, name, passwd):
        """Get HTML data from "name" host and return it.
           Try logon if it is requested.
        """

        response = opener.open(url_tmpl % name)
        response_data = response.read()
        logon_parser = LogonParser()
        # Workaround: Python 2.6.6, HTMLParser stop if there is "br" tag.
        logon_parser.feed(re.sub(BR_TAG_RE, "", response_data))
        if logon_parser.is_logon_page():
            response_data = self._logon_and_get_html(opener, name,
                                        logon_parser.get_form_data(passwd))

        return response_data

    def _get_xml_config(self, opener, url_tmpl, name, parser_class, passwd):
        """Get configuration from "name" host and return XML configuration
           data for the page specified by "url_tmpl".
        """

        try:
            response_data = self._get_html(opener, url_tmpl, name, passwd)
            parser = parser_class()

            # Workaround: Python 2.6.6, HTMLParser stop if there is "br" tag.
            parser.feed(re.sub(BR_TAG_RE, "", response_data))
            return parser.get_xml_config()
        except:
            self.exception("Fail _get_xml_config")
            # Ignore error
            pass

        return None

    def _append_crawled_data(self, name, passwd, landiskinfo):
        """Append crawled data to "landiskinfo"."""

        cookie_jar = cookielib.CookieJar()
        http_handler = urllib2.HTTPCookieProcessor(cookie_jar)
        opener = urllib2.build_opener(http_handler)

        target_info = (
                (DATETIME_CGI_URL_TMPL, DateTimeParser),
                (PROXY_CGI_URL_TMPL, ProxyParser),
                (SNMP_CGI_URL_TMPL, SnmpParser),
            )

        for (url_tmpl, parser_class) in target_info:
            xml_config = self._get_xml_config(opener, url_tmpl, name,
                                                parser_class, passwd)
            if xml_config != None:
                landiskinfo.append_child(xml_config)

        try:
            opener.open(LOGOFF_CGI_URL_TMPL % (name,))
        except:
            # Ignore error
            pass

    def _get_remote_landiskinfo(self, name, passwd):
        """Return the landiskinfo XML data of remote "name" landisk."""

        cmd_args = [GET_REMOTE_LANDISK_INFO_CMD, "-p", passwd, name]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        landiskinfo = XmlConfNode(proc.stdout.read())
        if proc.wait() != 0:
            raise ExecError(GET_REMOTE_LANDISK_INFO_CMD, proc.returncode)

        if self._is_hdl_xr_xv_series(landiskinfo):
            try:
                self._append_crawled_data(name, passwd, landiskinfo)
            except:
                self.exception("Fail _append_craled_data")
                # Ignore error
                pass

        return landiskinfo

    def _check_auth_to_remote_landiskinfo(self, name, passwd):
        """Check autorization to the remote "name" landisk."""

        cmd_args = [GET_REMOTE_LANDISK_INFO_CMD, "-t", "-p", passwd, name]
        CmdExec(cmd_args, stdout=NULL, stderr=NULL).execute()

    def _is_albg_running(self):
        """Return if albg_migration is running."""

        try:
            lock = FLock(config.LOCK_DIR, EXEC_LOCK_NAME, exclusive=True,
                            timeout=0)
            lock.lock()
            lock.unlock()
            # a lock file is not locked. albg_migration is not running.
            return False
        except:
            # a lock file is locked. albg_migration is running.
            return True

    def _get_status(self):
        """Return the current status."""

        if not os.path.exists(START_FILE_PATH):
            return STATUS_NO_EXEC

        status = STATUS_PROCESSING
        if os.path.exists(STATUS_FILE_PATH):
            with open(STATUS_FILE_PATH) as status_file:
                status = status_file.read().strip()

        if status == STATUS_PROCESSING and not self._is_albg_running():
            return STATUS_ABORT

        return status

    def _get_progress(self):
        """Return the data migration progress."""

        try:
            with open(PROGRESS_FILE_PATH) as progress_file:
                return progress_file.read().strip()
        except:
            return ""

    def _is_host_exists(self, hostname):
        """Check if "hostname" is exists."""

        ping_args = PING_ARG_BASE + [hostname]
        ping6_args = PING6_ARG_BASE + [hostname]

        if netut.is_inet4_addr(hostname):
            try:
                CmdExec(ping_args, stdout=NULL, stderr=NULL).execute()
                return True
            except:
                return False

        if netut.is_inet6_addr(hostname):
            try:
                CmdExec(ping6_args, stdout=NULL, stderr=NULL).execute()
                return True
            except:
                return False

        # "hostname" is no IPvX address. Try using ping and pin6.
        is_host_exists = False
        try:
            CmdExec(ping_args, stdout=NULL, stderr=NULL).execute()
            is_host_exists = True
        except:
            pass

        try:
            CmdExec(ping6_args, stdout=NULL, stderr=NULL).execute()
            is_host_exists = True
        except:
            pass

        return is_host_exists

    #
    # Request handlers
    #

    def _list_handler(self, entry):
        """Handle list request."""

        self._check_body_node(entry.get())
        try:
            sourcelist = XmlConfNode(SOURCELIST_NAME)
            hostname = self._get_hostname()
            for source_name in self._get_source_name_list():
                if source_name == hostname:
                    continue
                source = XmlConfNode(SOURCE_NAME)
                source.append_child(NAME_NAME, text=source_name)
                sourcelist.append_child(source)
            entry.get().append_child(sourcelist)
            entry.set_result(Results.OK)
        except:
            self.exception("get source fail.")
            entry.set_result(Results.LIST_ERROR)
        return entry

    def _query_handler(self, entry):
        """Handle query request."""

        self._check_body_node(entry.get())

        status = self._get_status()
        if status == STATUS_NO_EXEC:
            entry.get().append_child(XmlConfNode(STATUSINFO_NO_EXEC_XML))
            entry.set_result(Results.OK)
            return entry

        statusinfo = XmlConfNode(STATUSINFO_XML_TMPL % (status,))

        start_time = int(os.stat(START_FILE_PATH).st_mtime)
        statusinfo.get_node(STARTTIME_NAME).append_child(TIMESTAMP_NAME,
                        str(start_time))
        if status == STATUS_PROCESSING:
            if os.path.exists(PROGRESS_FILE_PATH):
                last_time = int(os.stat(PROGRESS_FILE_PATH).st_mtime)
                statusinfo.get_node(ELAPSEDTIME_NAME).set_text(
                            str(last_time - start_time))
                statusinfo.get_node(PROGRESS_NAME).set_text(
                                self._get_progress())
        elif os.path.exists(STATUS_FILE_PATH):
            statusinfo.get_node(ENDTIME_NAME).append_child(TIMESTAMP_NAME,
                            str(int(os.stat(STATUS_FILE_PATH).st_mtime)))

        entry.get().append_child(statusinfo)
        entry.set_result(Results.OK)
        return entry

    def _apply_handler(self, entry):
        """Handle apply request."""

        migration = entry.get()
        self._check_body_node(migration)
        try:
            name = migration.get_value(NAME_IN_JOB_PATH)
            passwd = migration.get_value(PASSWD_IN_JOB_PATH)
            self._check_auth_to_remote_landiskinfo(name, passwd)
        except:
            self.exception("Check Auth fail")
            entry.set_result(Results.AUTH_ERROR)
            return entry

        try:
            schedule = migration.get_node(SCHEDULE_PATH)
            if schedule.is_enabled():
                crond_ctrl.add_entry(CRON_ENTRY_NAME, EXEC_MIGRATION_CMD,
                        user="root",
                        month=schedule.get_value(MONTH_PATH, CRON_DEFAULT),
                        day=schedule.get_value(DAY_PATH, CRON_DEFAULT),
                        hour=schedule.get_value(HOUR_PATH, CRON_DEFAULT),
                        minutes=schedule.get_value(MINUTE_PATH, CRON_DEFAULT))
            else:
                crond_ctrl.delete_entry(CRON_ENTRY_NAME)

            entry.set_result(Results.OK)
        except:
            self.exception("apply fail.")
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def _get_handler(self, entry):
        """Handle get request."""

        migration = entry.get()
        self._check_body_node(migration)
        name = migration.get_value(NAME_PATH)
        passwd = migration.get_value(PASSWD_PATH)

        try:
            self._check_auth_to_remote_landiskinfo(name, passwd)
        except ExecError, err:
            if err.get_returncode() == EXIT_CODE_AUTH_ERROR:
                self.exception("get auth fail")
                entry.set_result(Results.AUTH_ERROR)
            else:
                self.exception("get connect fail")
                entry.set_result(Results.CONNECT_ERROR)
            return entry
        except:
            self.exception("get check_auth fail")
            entry.set_result(Results.AUTH_ERROR)
            return entry

        try:
            landiskinfo = self._get_remote_landiskinfo(name, passwd)
            entry.get().append_child(landiskinfo)
            entry.set_result(Results.OK)
        except:
            self.exception("get landisk info fail.")
            entry.set_result(Results.GET_ERROR)
        return entry

    def _execute_handler(self, entry):
        """Handle execute request."""

        self._check_body_node(entry.get(), entry.get_type())
        try:
            response_entry = self.send_entry(entry, ALBG_MIGRATION_PATH)
            if response_entry.get_result() != Results.OK:
                raise RequestError(response_entry.get_result())
            entry.set_result(Results.OK)
        except:
            self.exception("execute fail.")
            entry.set_result(Results.EXECUTE_ERROR)
        return entry

    def _check_handler(self, entry):
        """Handle check request."""

        self._check_body_node(entry.get(), entry.get_type())
        try:
            for hostname in entry.get().get_node_list(HOSTNAME_PATH):
                if self._is_host_exists(hostname.get_text()):
                    entry.set_result(Results.CHECK_ERROR)
                    return entry
        except:
            # Ignore error
            pass

        entry.set_result(Results.OK)
        return entry

    def _switch_handler(self, entry):
        """Handle switch request."""

        self._check_body_node(entry.get(), entry.get_type())
        try:
            response_entry = self.send_entry(entry, ALBG_MIGRATION_PATH)
            if response_entry.get_result() != Results.OK:
                raise RequestError(response_entry.get_result())
            entry.set_result(Results.OK)
        except:
            self.exception("switch fail.")
            entry.set_result(Results.SWITCH_ERROR)
        return entry

    #
    # Constructor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.LIST, self._list_handler)
        self.set_handler(ReqTypes.QUERY, self._query_handler)
        self.set_handler(ReqTypes.APPLY, self._apply_handler)
        self.set_handler(ReqTypes.GET, self._get_handler)
        self.set_handler(ReqTypes.EXECUTE, self._execute_handler)
        self.set_handler(ReqTypes.CHECK, self._check_handler)
        self.set_handler(ReqTypes.SWITCH, self._switch_handler)


#
# Main
#

if __name__ == "__main__":
    PalMigration(option=palcmd.PalCmdOption()).handle_request()
